#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include "config.h"
#include "types.h"
#define  PERMTABLES
#define  MAPTABLES
#include "access.h"
#include "util.h"
#include "dll.h"
#include "dllapps.h"
#include "globals.h"
#include "logfun.h"
#include "netio.h"
#include "readconfig.h"

/*
	convert a textual permission list into a bitmask
*/

INT evalPermission (CHAR *pnames)
{
    INT rtc = -1, perm = 0;
    perm_t *ptt;
    CHAR   *tok;

    for (tok = strtok(pnames,MKS); tok != NULL; tok = strtok(NULL,MKS))
        for (ptt = permtable; ptt->name != NULL; ptt++)
            if (!strcmp(tok, ptt->name)) {
                if (ptt->mask & PERM_CLEAR)
                    perm = 0;
                perm |= (ptt->mask & PERM_EMASK);
                rtc = 0;
            }
    return ( rtc ? rtc : perm );
}

/*
	Convert a permission mask INTo a comma separated text string
*/

CHAR *perm2names (INT permissionMask)
{
    INT i, count = 0;
    static CHAR permNameList[64];

    if (!permissionMask) 
        return permtable[0].name;

    permNameList[0] = '\0';

    for (i = 1; permtable[i].name != NULL; i++) {
        if (!(permtable[i].mask & PERM_ALIAS))
            if ((permissionMask & permtable[i].mask) == 
                (PERM_MASK      & permtable[i].mask)) {
                if (count++)
                    strcat(permNameList,",");
                strcat(permNameList,permtable[i].name);
            }
    }
    return permNameList;
}


VOID filterDump(FILE *dumpFILE)
{
    dlf_t *df;
    INT gotAny=0;

/*
	scan the filter list
*/

    fprintf (dumpFILE, "\nFilter declarations\n");
    if (filterListH != NULL) {
        for (df = NULL; (df = getNextFilter(filterListH, df)) != NULL;) {
            fprintf (dumpFILE, "    %d  %-30s  %-30s\n",
                     df->mark, df->rxfilter, df->txfilter);
            gotAny++;
        }
    }
    if (!gotAny)
        fprintf (dumpFILE,"        --none--\n");
}


VOID threadDump(FILE *dumpFILE)
{
    dla_t *da;
    struct in_addr ia;
    struct hostent *hp;
    INT gotAny=0;

/*
	scan the thread restriction list
*/

    fprintf (dumpFILE, "\nThread restrictions\n");
    if (threadListH != NULL) {
        for (da = NULL; (da = getNextACL(threadListH, da)) != NULL;) {
            fprintf (dumpFILE, "%d  ", da->acl.perm);
            if (!gotAny && (da->acl.mask == IPMASK_ANY)) {
                 fprintf (dumpFILE, "any\n");
                 gotAny++;
            }
            else if (da->acl.mask == IPMASK_HOST) {
                fprintf (dumpFILE, "%s ", "host");
                if (noResolv) {
                    ia.s_addr = da->acl.addr;
                    fprintf (dumpFILE, "%s\n", inet_ntoa(ia));
                }
                else {
                    hp = gethostbyaddr((CHAR *)&da->acl.addr,
                                       sizeof(in_addr_t),AF_INET);
                    if (hp == NULL) {
                        ia.s_addr = da->acl.addr;
                        fprintf (dumpFILE, "%s\n", inet_ntoa(ia));
                    }
                    else
                        fprintf (dumpFILE, "%s\n", hp->h_name);
                }
            }
            else {
                ia.s_addr = (da->acl.addr & da->acl.mask);
                fprintf (dumpFILE, "%s ", inet_ntoa(ia));
                ia.s_addr = da->acl.mask;
                fprintf (dumpFILE, "%s\n",inet_ntoa(ia));
            }
        }
    }
    if (!gotAny)
        fprintf (dumpFILE,"        --none--\n");
}


VOID accessDump(FILE *dumpFILE)
{
    dll_t *save, *sp;
    dlr_t *dr;
    dla_t *da;
    struct in_addr ia;
    struct hostent *hp;
    INT dummy = 0, allLists = 0, allEntries, gotAny;

    if ((save = initChain(NULL)) == NULL) {
        errNoDynmem(NULL);
        return;
    }

/*
	scan the restriction list and the accesslist chain
	to obtain all allocated list numbers
*/

    if (restrictListH != NULL) {
        for (dr = NULL; (dr = getNextRe(restrictListH, dr)) != NULL;) {

            for (sp = NULL; (sp = getNext(save, sp)) != NULL;)
                if (sp->mark == dr->mark) break;

            if (sp == NULL) {
                if (addPlain(save, dr->mark, &dummy, sizeof(dummy))) {
                    errNoDynmem(NULL);
                    return;
                }
                allLists++;
            }
        }
    }

    if (accessListH != NULL) {
        for (da = NULL; (da = getNextACL(accessListH, da)) != NULL;) {

            for (sp = NULL; (sp = getNext(save, sp)) != NULL;)
                if (sp->mark == da->mark) break;

            if (sp == NULL) {
                if (addPlain(save, da->mark, &dummy, sizeof(dummy))) {
                    errNoDynmem(NULL);
                    return;
                }
                allLists++;
            }
        }
    }

    fprintf (dumpFILE, "\nTotal number of access lists allocated: %d\n", 
             allLists);

    if (!allLists)
        return;

/*
	Now that we know all the lists, sort them by mark and dump them
*/

    sortByMark(save, DLL_SORTASC);

    for (sp = NULL; (sp = getNext(save, sp)) != NULL;)  {

        fprintf (dumpFILE, "\nAccess list %d\n",sp->mark);

        for (da = NULL, gotAny = 0, allEntries = 0; 
             (da = getNextACL(accessListH, da)) != NULL;)
            if (sp->mark == da->mark) {
                allEntries++;
                fprintf (dumpFILE, "    %-16s ", 
                         perm2names(da->acl.perm));
                if (!gotAny && (da->acl.mask == IPMASK_ANY)) {
                     fprintf (dumpFILE, "any\n");
                     gotAny++;
                }
                else if (da->acl.mask == IPMASK_HOST) {
                    fprintf (dumpFILE, "%s ", "host");
                    if (noResolv) {
                        ia.s_addr = da->acl.addr;
                        fprintf (dumpFILE, "%s\n", inet_ntoa(ia));
                    }
                    else {
                        hp = gethostbyaddr((CHAR *)&da->acl.addr,
                                           sizeof(in_addr_t),AF_INET);
                        if (hp == NULL) {
                            ia.s_addr = da->acl.addr;
                            fprintf (dumpFILE, "%s\n", inet_ntoa(ia));
                        }
                        else
                            fprintf (dumpFILE, "%s\n", hp->h_name);
                    }
                }

                else {
                    ia.s_addr = (da->acl.addr & da->acl.mask);
                    fprintf (dumpFILE, "%s ", inet_ntoa(ia));
                    ia.s_addr = da->acl.mask;
                    fprintf (dumpFILE, "%s\n",inet_ntoa(ia));
                }
            }

        if (allEntries) {
            if (!gotAny)
                fprintf (dumpFILE, "    %-16s any\n", "deny");
        }
        else                       
            fprintf (dumpFILE, "    %-16s any\n", "read,write");

        dr = getNextRe(restrictListH, NULL);

        if (dr != NULL) {
            int gotAnyR = 0;
            fprintf (dumpFILE, "\n    File restrictions for access list %d\n",
                     sp->mark);
            do {
                if (sp->mark == dr->mark) {
                    fprintf(dumpFILE, "        %-30s  %d\n",dr->raw,dr->filter);
                    gotAnyR++;
                }
            } while ((dr = getNextRe(restrictListH, dr)) != NULL);
            if (!gotAnyR)
                fprintf (dumpFILE,"        --none--\n");
        }

        dr = getNextRe(mapListH, NULL);

        if (dr != NULL) {
            int gotAnyR = 0;
            fprintf (dumpFILE, "\n    File name maps for access list %d\n",
                     sp->mark);
            do {
                if (sp->mark == dr->mark) {
                    fprintf(dumpFILE, "        %-30s  %-30s\n",dr->raw,dr->map);
                    gotAnyR++;
                }
            } while ((dr = getNextRe(mapListH, dr)) != NULL);
            if (!gotAnyR)
                fprintf (dumpFILE,"        --none--\n");
        }
    }
}

/*
	check if the system will give access to the file 'stat'ed
	for the current access mode
*/

static INT haveAccess (struct stat *sp, INT reqtype)
{
    if (reqtype == RT_RRQ) {
        if ( (sp->st_mode & S_IROTH) ||
            ((sp->st_mode & S_IRUSR) && (sp->st_uid == geteuid())) ||
            ((sp->st_mode & S_IRGRP) && (sp->st_gid == getegid())))
            return 1;
    }
    else if (reqtype == RT_WRQ) {
        if ( (sp->st_mode & S_IWOTH) ||
            ((sp->st_mode & S_IWUSR) && (sp->st_uid == geteuid())) ||
            ((sp->st_mode & S_IWGRP) && (sp->st_gid == getegid())))
            return 1;
    }
    return 0;
}

/*
	Rules:
	1.1)	In absence of any access list, grant read/write
	1.2)	find the client in the main access list.
	1.3)	if no matching entry exists at all, deny access anyway.

	2.1)	scan the file map list for a match of the requested file.
	2.2)	if found, replace the requested file name with the translated
		file name and proceed.
	
	3.1)	scan all restrict lists that match the number of the
		currently valid access list for a match of the requested file.
	3.2)	if no restrict list for that access list exists,
		grant access according to the access list entry.
	3.2)	if found a match, grant access according to the access 
		list entry too.
	3.3)	if no match found, proceed to the next access list entry.
*/

static INT checkACL (tftpR_t *request)
{
    dla_t *ap;
    dlr_t *rp;
    dlf_t *fp;
    dlr_t *mp;
    INT pmask;
    in_addr_t peer;

    peer  = (in_addr_t)request->reqPeer.sin_addr.s_addr;
    pmask = (request->reqType == RT_RRQ) ? PERM_READ :
                                           PERM_WRITE|PERM_CREATE|PERM_APPEND;
/*
	No access lists specified, fall back to traditional tftpd
*/
    if (isEmptyChain((dll_t *) accessListH)) {
        LOG_MSG2(LOG_WARNING, "ACL empty, permission is read,write");
        return (PERM_READ | PERM_WRITE);
    }

/*
	scan the access lists
*/
    for (ap = NULL; (ap = findACL(accessListH, ap, 0, peer)) != NULL; ) { 

        if (((ap->acl.perm & PERM_EMASK) == PERM_DENY)) {
            LOG_MSG4(LOG_DEBUG, "ACL #%d denies explicitely", ap->mark);
            return PERM_NONE;
        }

        if ((ap->acl.perm & pmask)) {
            LOG_MSG4(LOG_DEBUG, "ACL #%d permits %s", 
                     ap->mark, perm2names(ap->acl.perm));
        }

/*
	scan the file map lists to see if we need a name translation
*/
        if ((mp = findRePattern(mapListH, NULL, 
                                ap->mark, request->reqFile)) != NULL) {
            request->realFile = xlateFileMap(request->reqFile, mp->map,
                                             &request->reqPeer);
            LOG_MSG4(LOG_DEBUG, "Map #%d from '%s' to '%s'",
                                mp->mark, request->reqFile, 
                                request->realFile);
        }
        else {
            request->realFile = request->reqFile;
        }

/*
	scan the restrict lists
*/
        if ((rp = findRePattern(restrictListH, NULL,
                                ap->mark,request->realFile)) != NULL) {
            LOG_MSG4(LOG_DEBUG, "Restriction '%s' for access list #%d"
                                " grants '%s'",
                                rp->raw, ap->mark, request->realFile);
        }
        else {
            if ((rp = getReByMark(restrictListH, NULL, ap->mark)) == NULL) {
                LOG_MSG4(LOG_DEBUG, "No restrictions for access list #%d",
                                    ap->mark);
            }
            else {
                LOG_MSG4(LOG_DEBUG, "Restrictions for access list #%d"
                                    " deny '%s'",
                                     ap->mark, request->realFile);
                continue;
            }
        }

/*
	now, that the access is granted anyway,
	check here if we are intended to use i/o filtering
*/
        if ((rp != NULL) && (needFilter = rp->filter)) {
            if ((fp = getFilterByMark(filterListH, NULL, 
                                      needFilter)) != NULL) {
                if (request->reqType == RT_RRQ) {
                    if (fp->rxfilter != NULL && strlen(fp->rxfilter) != 0) {
                        request->filter = fp->rxfilter;
                    }
                    else {
                        needFilter = 0;
                    }
                }
                else {
                    if (fp->txfilter != NULL && strlen(fp->txfilter) != 0) {
                        request->filter = fp->txfilter;
                    }
                    else {
                        needFilter = 0;
                    }
                }
            } 
            else {
                needFilter = 0;
            }
            if (needFilter)
                LOG_MSG4(LOG_DEBUG, "Applying i/o filter #%d '%s'",
                         fp->mark, request->filter);
        }
        return (ap->acl.perm & pmask);
    }
    LOG_MSG4(LOG_DEBUG, "No matching access list found");
    return PERM_NONE;
}

/*
	Validate file access on basis of request (transfer) type, 
	requested file and the peer's ip address.
	returns 0 and a FILE* to the opened requestd file
	on success or an TFTP error code else.
*/

INT validateAccess (tftpR_t *request, INT fake)
{
    struct stat stbuf;
    enum openModes { READ, WRITE, CREATE } openMode;
    INT perm, ds, cmode, omask; 
    CHAR *filename, *fname;

/*
	First of all, evaluate the access lists
*/

    perm = checkACL (request);

    filename = request->realFile;

/*
	Rule 1:

	We deny all requests for a relative path, if we haven't got
	either a virtual root or a default directory path
*/
    if (filename != NULL && *filename != '/') {
        if (DefaultDirectory == NULL && RootDirectory == NULL) {
            LOG_MSG2 (LOG_WARNING, "access denied%s",
                      SafeLog ? "" : ": illegal relative file path");
            return PX_ACCESS;
        }
        if (DefaultDirectory != NULL) {
            CHAR *temp;
            temp = realPath (DefaultDirectory, filename);
            if (temp == NULL) {
                errNoDynmem("Cannot create realPath");
                return PX_NOSPACE;
            }
            filename = alloca(strlen(temp)+1);
            if (filename == NULL) {
                errNoDynmem("Cannot copy realPath");
                return PX_NOSPACE;
            }
            strcpy (filename, temp);
            free (temp);
        }
    }
/*
	Rule 2:

	The requested operation must match the current access lists
*/

    if ((request->reqType == RT_RRQ) && !(perm & PERM_READ)) {
        LOG_MSG2 (LOG_WARNING, "%saccess denied%s",
                  SafeLog ? "" : "read ",
                  SafeLog ? "" : " due to ACL");
        return PX_ACCESS;
    }
    else if ((request->reqType == RT_WRQ) && 
             !(perm & PERM_WRCRAP)) {
        LOG_MSG2 (LOG_WARNING, "%saccess denied%s",
                  SafeLog ? "" : "write/append/create ",
                  SafeLog ? "" : " due to ACL");
        return PX_ACCESS;
    }
    else {
        LOG_MSG4 (LOG_DEBUG, "ACL: %s%saccess possible",
                  SafeLog ? "" : perm2names(perm),
                  SafeLog ? "" : " ");
    }

/*
	Rule 3:

	If a new file is to be created, the parent directory must be writable
*/
    memset ((CHAR *) &stbuf, 0, sizeof (stbuf));

    if (stat (filename, &stbuf) < 0) {
        if (errno == ENOENT) {
            if ((perm & PERM_CREATE)) {
                INT rtc;
                CHAR *dirname;
                if ((dirname = strdup(filename)) == NULL) {
                    errNoDynmem(NULL);
                    return PX_NOSPACE;
                }

                rtc = dpstat (dirname, &stbuf);

                if (rtc < 0) {
                    LOG_MSG2 (LOG_WARNING, "directory %s not found", dirname);
                    free (dirname);
                    return PX_ACCESS;
                }
                else {
                    free (dirname);
                    if (!haveAccess(&stbuf, RT_WRQ)) {
                        LOG_MSG2 (LOG_WARNING, "%saccess denied%s",
                                  SafeLog ? "" : "create ",
                                  SafeLog ? "" : " due to system");
                        return PX_ACCESS;
                    }
                    openMode = CREATE;
                } 
            }
/* 
	read/write access to non-existent file 
*/
            else {
                if (SafeLog) {
                    LOG_MSG2 (LOG_WARNING, "access denied");
                }
                else {
                    LOG_MSG2 (LOG_WARNING, "%s access denied: %m",
                              (request->reqType == RT_RRQ) ? "read" : "write");
                return PX_NOTFOUND;
                }
            }
        }
/*
	all other stat() errors other than ENOENT
*/
        else {
            LOG_MSG2 (LOG_WARNING, "Cannot get file status: %m");
            return PX_UNDEF;
        }
    }
/* 
	file status checks o.k., i.e. the file or directory exists
*/
    else {
/*
	create-only acces to existing file
*/
        if ((perm & PERM_WRCRAP) == PERM_CREATE) {
            LOG_MSG2 (LOG_WARNING, "%saccess denied%s",
                      SafeLog ? "" : "create only ",
                      SafeLog ? "" : ": file exists");
            return PX_ACCESS;
        }
/*
	Rule 4:

        Allow access only to regular files
*/
        if (!S_ISREG(stbuf.st_mode)) {	/* allow regular files only */
            LOG_MSG2 (LOG_WARNING, "access denied%s",
                      SafeLog ? "" : ": not a regular file");
            return PX_ACCESS;
        }
/*
	Rule 5:

	Deny access to our own config file anyway
*/
        if (isConfigFile(&stbuf)) {	/* inhibit config file access */
            LOG_MSG2 (LOG_WARNING, "access denied%s",
                      SafeLog ? "" : " to config file ");
            return PX_ACCESS;
        }
/*
	Rule 6:

	The system file modes must give us the appropriate (read/write)
	access to the file in question.
*/
        if (!haveAccess(&stbuf, request->reqType)) {
            LOG_MSG2 (LOG_WARNING, "%saccess denied%s%s%s",
                      SafeLog ? "" : (request->reqType==RT_RRQ) ? "read " : "write ",
                      SafeLog ? "" : " due to system permissions (",
                      SafeLog ? "" : fmtstat(&stbuf),
                      SafeLog ? "" : ")");
            return PX_ACCESS;
        }
        if (request->reqType == RT_RRQ) 
            openMode = READ;
        else
            openMode = WRITE;
    }

/*
	at this point access to the requested file is granted
	try to open / create the file
*/

    if (fake)
        return 0;
                
    switch (openMode) {

        case READ:
            ds = open (filename, O_RDONLY);

            if (ds >= 0) {
                if (needFilter) {
                    if (forkFilter (request->filter, request->reqType, &ds)) {
                        close (ds);
                        return PX_ACCESS;
                    }
                }
                request->reqFd = fdopen(ds, "r");
            }
            else
                request->reqFd = NULL;
            break;

        case WRITE:
            if (perm & PERM_APPEND) {
                if (needFilter) {
                    LOG_MSG1 (LOG_ERR, "Cannot APPEND using a filter");
                    return PX_ACCESS;
                }

                ds = open (filename, O_WRONLY | O_APPEND);

                if (ds >= 0)
                    request->reqFd = fdopen(ds, "a");
                else
                    request->reqFd = NULL;
            }
            else {
                ds = open (filename, O_WRONLY | O_TRUNC);

                if (ds >= 0) {
                    if (needFilter) {
                        if (forkFilter (request->filter, 
                                        request->reqType, &ds)) {
                            close (ds);
                            return PX_ACCESS;
                        }
                    }
                    request->reqFd = fdopen(ds, "w");
                }
                else
                    request->reqFd = NULL;
            }
            break;
  
        case CREATE:
            cmode = S_IRUSR | S_IRGRP | S_IROTH;
            if (perm & (PERM_WRITE|PERM_APPEND)) cmode |= (S_IWUSR | S_IWGRP);
            omask = umask (0);

            LOG_MSG5 (LOG_DEBUG, "Create mode: %o", cmode);
            
            ds = open (filename, O_WRONLY | O_CREAT | O_TRUNC | O_EXCL, cmode);

            if (ds >= 0) {
                (VOID) umask (omask);

                if (needFilter) {
                    if (forkFilter (request->filter, 
                                    request->reqType, &ds)) {
                        close (ds);
                        return PX_ACCESS;
                    }
                }
                request->reqFd = fdopen(ds, "w");
            }
            else
                request->reqFd = NULL;
            break;
    }

    if (request->reqFd == NULL) {
        LOG_MSG2 (LOG_WARNING, "fd/open (%s) fails for %s: %m", 
                  openMode == READ ? "READ" : openMode == CREATE ?
                  "CREATE" : (perm & PERM_APPEND) ? "APPEND" : "WRITE",
                  filename, strerror(errno));
        return PX_ACCESS;
    }
    fname = baseName(request->filter);
    LOG_MSG4 (LOG_DEBUG, "File '%s' opened for %s%s%s%s", filename,
              openMode == READ ? "READ" : openMode == CREATE ?
              "CREATE" : (perm & PERM_APPEND) ? "APPEND" : "WRITE",
              needFilter ? " via filter '" : "", 
              needFilter ? fname : "",
              needFilter ? "'" : "");
    if (fname != NULL)
        free (fname);
    return (0);
}

/*
	check thread restrictions
*/

INT checkThreads(tftpR_t *req, dlp_t **np)
{
    dla_t *da = NULL;
    dlp_t *dp;
    INT count;
    in_addr_t peer;

/*
	scan the thread restriction list for a matching entry.
	if nothing matching found, we assume -no- restriction
*/

    peer = (in_addr_t)req->reqPeer.sin_addr.s_addr;
    if ((da = findACL(threadListH, da, 0, peer)) == NULL)
        return -1;
/*
	now we know, how many threads of this kind are allowed.
	we're going to scan the process list for entries of this type
*/
    count = da->acl.perm;

    for (dp = NULL; (dp = getNextPid(procListH, dp)) != NULL; )
        if (dp->mark == da->mark) {
            count--;
            if (count <= 0)
                break;
        }

    if (count <= 0) {
        struct in_addr ia;
        ia.s_addr = (da->acl.addr & da->acl.mask);
        LOG_MSG2 (LOG_WARNING, "Limit %d reached for %s",
                  da->acl.perm, inet_ntoa(ia));
        return 0;
    }

    if (addPid(procListH, da->mark, 0, 0, 0, 0, 0, &dp)) {
        errNoDynmem(NULL);
        exit(1);
    }

    if (np != NULL)
        *np = dp;

    return count;
}

VOID dumpProcList(VOID)
{
    INT count;

    if ((count = getPidCount(procListH))) {
        dlp_t *dp;
        CHAR *dbuf;

        if ((dbuf = (CHAR *)alloca((count*11)+1)) == NULL) {
            errNoDynmem(NULL);
            return;
        }

        *dbuf = '\0';

        for (dp = NULL; (dp = getNextPid(procListH, dp)) != NULL; )
            sprintf((dbuf+strlen(dbuf)), "%u ", dp->pid);

        LOG_MSG4 (LOG_DEBUG,"%d active subprocesses: %s", count, dbuf);
    }
    else 
        LOG_MSG4 (LOG_DEBUG,"No active subprocesses");
}

INT getProcNum(VOID)
{
    return getPidCount(procListH);
}

#ifdef TEST

INT main (INT argc, CHAR **argv)
{
    INT i, rtc;
    CHAR *arg[3], *lx, line[100], pt[64];
    tftpR_t req;

    evalCommandLine(argc, argv, 0);
    spMode   = 1;
    MsgLevel    = 5;
    MsgOverride = 5;
    MaxSyslog   = 0;
    MslOverride = 0;
    CheckOnly   = 1;
    startLog();
    rtc = readConfigFile (ConfigFile);
    fprintf (stderr, "Status of config file processing %d\n",rtc);
    changeIdent();

    do {
        fprintf(stderr,"access-test>");
        fflush(stderr);
        if (fgets(line, sizeof(line), stdin) == NULL)
            break;
        fprintf(stderr,"\n");

        for (i=0, lx=line; i<3; i++, lx=NULL) {
            if((arg[i]= strtok(lx, ", \t\n")) == NULL) {
                fprintf(stderr,"Need 3 arguments: <r|w> ip-address file\n");
                break;
            }
        }
        if (i < 3) continue;
        memset (&req, 0, sizeof(req));
        req.reqType                 = (*arg[0] == 'r') ? RT_RRQ : RT_WRQ;
        req.reqPeer.sin_addr.s_addr = xinet_addr(arg[1], 0);
        req.reqFile                 = arg[2];

        if(!validateAccess(&req, 1)) {
            i = checkThreads(&req, NULL);
            if (i >= 0)
                sprintf(pt,"%d",i);
            else
                strcpy(pt, "unlimited");
            fprintf(stderr,"Access granted for %s parallel transfers\n",pt);
        }
    } while (1);
    fprintf(stderr,"\n");
    return 0;
}
#endif /* TEST */
